
#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.functional_exception import FunctionalException
import re
from holado.common.handlers.object import Object, DeleteableObject
from holado_core.common.tools.tools import Tools
from holado.holado_config import Config
from contextlib import contextmanager
from typing import NamedTuple
import threading
from holado.common.tools.gc_manager import GcManager
from holado_python.standard_library.typing import Typing
from holado.common.context.session_context import SessionContext

logger = logging.getLogger(__name__)
# logger_update = logging.getLogger(__name__ + ".update")


class VariableManager(DeleteableObject):
    """
    Manage variables as couples of (name, value).
    
    Variable names can contain any character except: ".", "["
    Variable names can end by any dynamic suffix symbol, the symbol will be replaced by the associated suffix.
    """

    def __init__(self, parent=None):
        super().__init__("variable manager")
        
        self.__parent = parent
        self.__dynamic_text_manager = None
        self.__unique_value_manager = None
        self.__variables = {}
        
        # Manage temporary variables
        self.__temporary_variables_lock = threading.Lock()
        self.__temporary_variables = {}
        
        # Manage variable update log file
        self.__variable_update_log_file_path = None
        self.__is_entered_in_variable_update_log_file = False
        self.__logger_update = logging.getLogger(__name__ + f".update.{id(self)}")
    
    def initialize(self, dynamic_text_manager, unique_value_manager, variable_update_log_file_path=None):
        """ Initialize variable manager
        @param dynamic_text_manager Dynamic text manager instance to use
        @param unique_value_manager Unique value manager instance to use
        @param variable_update_log_filename If defined, log all variable updates in given file in addition to normal logs
        """
        self.__dynamic_text_manager = dynamic_text_manager
        self.__unique_value_manager = unique_value_manager
        
        # Manage variable update log file
        self.__variable_update_log_file_path = variable_update_log_file_path
    
    @property
    def parent(self):
        return self.__parent
    
    def _delete_object(self):
        # Delete all variables
        names = set(self.__variables.keys())
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Removing {len(names)} variables: {names}")
        for index, name in enumerate(names):
            if name in self.__variables:
                value = self.__variables[name]
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Removing variable {index+1}/{len(names)}: '{name}' (type: {Typing.get_object_class_fullname(value)})")
                self.unregister_variable(name, through_parent=False, delete_value=True)
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Finished to remove {len(names)} variables: {names}")
                
        # # Clear variables
        # self.__variables.clear()
        
        # Leave variable update log file
        self.__leave_variable_update_log_file()
    
    def __enter_variable_update_log_file(self):
        if self.__variable_update_log_file_path is not None and not self.__is_entered_in_variable_update_log_file:
            SessionContext.instance().log_manager.enter_log_file_for_logger(self.__logger_update, self.__variable_update_log_file_path, use_format_short=True)
            self.__is_entered_in_variable_update_log_file = True
    
    def __leave_variable_update_log_file(self):
        if self.__variable_update_log_file_path is not None and self.__is_entered_in_variable_update_log_file:
            SessionContext.instance().log_manager.leave_log_file_for_logger(self.__logger_update)
            self.__is_entered_in_variable_update_log_file = False
    
    @contextmanager
    def temporary_variables(self):
        from holado_multitask.multitasking.multitask_manager import MultitaskManager
        
        uid = MultitaskManager.get_thread_uid()
        with self.__temporary_variables_lock:
            if uid not in self.__temporary_variables:
                self.__temporary_variables[uid] = NamedTuple("TempVars", counter=int, variable_names=list)
                self.__temporary_variables[uid].counter = 0
                self.__temporary_variables[uid].variable_names = set()
            self.__temporary_variables[uid].counter += 1
            
        try:
            yield None
        finally:
            with self.__temporary_variables_lock:
                self.__temporary_variables[uid].counter -= 1
                if self.__temporary_variables[uid].counter == 0:
                    logger.info(f"Unregistering temporary variables: {self.__temporary_variables[uid].variable_names}")
                    for var_name in self.__temporary_variables[uid].variable_names:
                        if self.exists_variable(var_name, through_parent=False):
                            self.unregister_variable(var_name, through_parent=False, delete_value=True)
                    del self.__temporary_variables[uid]
                    GcManager.collect()
        
        
    def new_local_variable_name(self, sub_name=None):
        if sub_name is not None:
            return f"__VAR_{sub_name}_{self.__unique_value_manager.new_string(do_log=False)}__"
        else:
            return f"__VAR{self.__unique_value_manager.new_string(do_log=False)}__"
    
    def new_variable_name(self):
        res = None
        while True:
            res = self.new_local_variable_name()
            if not self.exists_variable(variable_name=res):
                break
        return res
    
    def evaluate_variable_name(self, var_name):
        if var_name.endswith(Config.DYNAMIC_SYMBOL):
            return self.__dynamic_text_manager.get(var_name[:-1])
        elif var_name.endswith(Config.THREAD_DYNAMIC_SYMBOL):
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            return self.__dynamic_text_manager.get(var_name[:-1], scope=MultitaskManager.get_thread_id())
        elif var_name.endswith(Config.UNIQUE_SYMBOL):
            return var_name[:-1] + self.__unique_value_manager.new_string(padding_length=Config.unique_string_padding_length)
        else:
            return var_name
    
    def register_variable(self, variable_name, value, accept_expression=True, log_level=logging.INFO):
        from holado_multitask.multitasking.multitask_manager import MultitaskManager
        
        self.__enter_variable_update_log_file()
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Setting variable: {variable_name}=[{value}] (type: {Typing.get_object_class_fullname(value)})...")
        
        names = variable_name.rsplit('.', 1)
        if accept_expression and len(names) > 1:
            if not self.exists_variable(expression=names[0]):
                self.register_variable(names[0], Object())
            var = self.get_value(names[0], through_patent=False)
            setattr(var, names[1], value)
            if Tools.do_log(self.__logger_update, log_level):
                self.__logger_update.log(log_level, f"Set variable expression: {variable_name}=[{value}] (type: {Typing.get_object_class_fullname(value)})")
        else:
            var_name = self.evaluate_variable_name(variable_name)
            self.__variables[var_name] = value
            if Tools.do_log(self.__logger_update, log_level):
                self.__logger_update.log(log_level, f"Set variable: {var_name}=[{value}] (type: {Typing.get_object_class_fullname(value)})")
                
            # Manage temporary variables
            uid = MultitaskManager.get_thread_uid()
            with self.__temporary_variables_lock:
                if uid in self.__temporary_variables:
                    self.__temporary_variables[uid].variable_names.add(var_name)
                
    def unregister_variable(self, variable_name, through_parent=False, delete_value=False):
        self.__enter_variable_update_log_file()
        
        var_name = self.evaluate_variable_name(variable_name)
        if self.exists_variable(var_name, through_parent=False):
            value = self.__variables[var_name]
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Unregistering variable '{var_name}' (type: {Typing.get_object_class_fullname(value)})")
                
            if delete_value and isinstance(value, DeleteableObject):
                try:
                    value.delete_object()
                except Exception:
                    logger.exception(f"Catched exception while deleting object '{var_name}' (type: {Typing.get_object_class_fullname(value)})")
                else:
                    if Tools.do_log(self.__logger_update, logging.DEBUG):
                        self.__logger_update.debug(f"Deleted variable '{var_name}' (type: {Typing.get_object_class_fullname(value)})")
            
            del self.__variables[var_name]
        elif through_parent and self.parent:
            return self.parent.unregister_variable(variable_name, through_parent=through_parent, delete_value=delete_value)
        else:
            raise FunctionalException(f"Variable '{var_name}' has not been found. Missing initialisation ?")
        
    def pop_variable(self, variable_name):
        res = self.get_variable_value(variable_name, through_parent=False)
        self.unregister_variable(variable_name, through_parent=False, delete_value=False)
        return res
    
    def get_variable_value(self, variable_name, attributes = [], through_parent=True):
        # Get variable value
        var_name = self.evaluate_variable_name(variable_name)
        if self.exists_variable(var_name, through_parent=False):
            res = self.__variables[var_name]
        elif through_parent and self.parent:
            return self.parent.get_variable_value(variable_name, attributes)
        else:
            raise FunctionalException("Variable '{}' has not been found. Missing initialisation ?".format(variable_name))
        
        # If needed, call attributes
        for attr in attributes:
            res = getattr(res, attr)
        
        return res

    def get_value(self, expression, through_patent=True):
        res = None
        
        if self.is_variable_expression(expression, through_parent=through_patent):
            try:
                res = self.eval_variable_expression(expression, through_parent=through_patent)
            except (SyntaxError, NameError):
                raise FunctionalException("Expression '{}' is not evaluable. Is an initialization missing ?".format(expression))
        else:
            res = self.get_variable_value(expression, through_parent=through_patent)
        
        return res

    def set_value(self, expression, value, through_parent=True):
        if self.is_variable_expression(expression, through_parent=through_parent):
            try:
                self.exec_set_variable_expression(expression, value, through_parent=through_parent)
            except (SyntaxError, NameError):
                raise FunctionalException("Expression '{}' is not evaluable. Is an initialization missing ?".format(expression))
        else:
            self.register_variable(expression, value)

    def exists_variable(self, variable_name=None, expression=None, through_parent=True):
        if expression is not None:
            expression = expression.strip()
            m_operator = re.search("[.[]", expression)
            if m_operator:
                # Return directly False if found operator is not an operator
                # EKL: Why this check is done ???
                if m_operator.group(0) not in ['.', '[']:
                    return False
                # Begin by looking for variable existance
                variable_name = expression[:m_operator.start()]
            else:
                variable_name = expression
        
        var_name = self.evaluate_variable_name(variable_name)
        res = var_name in self.__variables
        
        # For expression, look if sub variable exist
        if res and expression is not None and variable_name != expression:
            try:
                _ = self.eval_variable_expression(expression, through_parent=False)
            except:
                res = False
            else:
                res = True
        
        if not res and through_parent and self.parent:
            if expression is not None:
                res = self.parent.exists_variable(expression=expression)
            else:
                res = self.parent.exists_variable(variable_name=variable_name)
        return res
    
    def is_variable_expression(self, expression, through_parent=True):
        try:
            var_name, expr_rest = self.__extract_expression_name(expression)
            return len(var_name) > 0 and self.exists_variable(var_name, through_parent=through_parent) and len(expr_rest) > 0
        except SyntaxError:
            return False
    
    def eval_variable_expression(self, expression, through_parent=True):
        var_name, expr_rest = self.__extract_expression_name(expression)
        var_value = self.get_variable_value(var_name, through_parent=through_parent)
        new_var_name = self.new_local_variable_name(sub_name=self.__regular_variable_name(var_name))
        locals()[new_var_name] = var_value
        
        new_expr = f"{new_var_name}{expr_rest}"
        try:
            return eval(new_expr)
        except Exception as exc:
            try:
                var_repr = Tools.represent_object(var_value)
            except:
                var_repr = str(var_value)
            Tools.raise_same_exception_type(exc, f"{exc}\nVariable '{var_name}': {var_repr}", add_from=True)
    
    def exec_set_variable_expression(self, expression, value, through_parent=True, log_level=logging.INFO):
        self.__enter_variable_update_log_file()
        
        var_name, expr_rest = self.__extract_expression_name(expression)
        var_value = self.get_variable_value(var_name, through_parent=through_parent)
        new_var_name = self.new_local_variable_name(sub_name=self.__regular_variable_name(var_name))
        locals()[new_var_name] = var_value
        
        new_var_value = self.new_local_variable_name(sub_name='value')
        locals()[new_var_value] = value
        
        new_expr = f"{new_var_name}{expr_rest}={new_var_value}"
        try:
            exec(new_expr)
        except Exception as exc:
            try:
                var_repr = Tools.represent_object(var_value)
            except:
                var_repr = str(var_value)
            # Tools.raise_same_exception_type(exc, f"{exc}\nVariable '{var_name}': {var_repr}", add_from=True)
            raise FunctionalException(f"Failed to set value in variable expression (by exec): {expression}=[{value}] (type: {Typing.get_object_class_fullname(value)})\nVariable '{var_name}': {var_repr}") from exc
        else:
            if Tools.do_log(self.__logger_update, log_level):
                self.__logger_update.log(log_level, f"Set value in variable expression (by exec): {expression}=[{value}] (type: {Typing.get_object_class_fullname(value)})")
    
    def __regular_variable_name(self, name):
        return "".join([c if re.match(r'\w', c) else '_' for c in name])
    
    # def set_variable_expression(self, expression, value):
    #     var_name, expr_rest = self.__extract_expression_name(expression)
    #     if len(expr_rest) == 0:
    #         self.register_variable(expression, value)
    #     else:
    #         var_value = self.get_variable_value(var_name)
    #
    #     while len(expr_rest) > 0:
    #

    
    def __extract_expression_name(self, expression):
        regex_name = re.compile("^([^.[]+)(.*)")
        m = regex_name.match(expression)
        if not m:
            raise SyntaxError(f"Expression [{expression}] is not a variable expression")
        var_name, expr_rest = m.group(1), m.group(2)
        var_name = self.evaluate_variable_name(var_name)
        return var_name, expr_rest
        
        
        
        
        




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado.common.handlers.undefined import to_be_defined
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools

try:
    import xmltodict  # @UnresolvedImport
    with_xmltodict = True
except:
    with_xmltodict = False

try:
    from lxml import etree  # @UnresolvedImport
    with_lxml = True
except:
    with_lxml = False



class XML2Dict(object):
    """
    Manage conversion of XML <-> dict
    """
    
    
    @classmethod
    def parse(cls, content, force=False, **parse_kwargs):
        """ Parse content to dict, in libraries order and if available:
          - xmltodict,
          - lxml.
        If parse has failed and 'force' is True, a parse is tried with lxml (if available) in recovering mode, in case of XML files well-formed but with "invalid" characters
        """
        res = to_be_defined
        
        try:
            if with_xmltodict:
                kwargs = parse_kwargs['xmltodict'] if 'xmltodict' in parse_kwargs else {}
                res = cls.parse_with_xmltodict(content, **kwargs)
            elif with_lxml:
                kwargs = parse_kwargs['lxml'] if 'lxml' in parse_kwargs else {}
                res = cls.parse_with_lxml(content, **kwargs)
            else:
                raise TechnicalException(f"No XML library is available in (xmltodict, lxml)")
        except:
            # Parse with lxml in recovering mode, in case of XML files with "invalid" characters
            if force and with_lxml:
                kwargs = parse_kwargs['lxml'] if 'lxml' in parse_kwargs else {}
                kwargs['recover'] = True
                res = cls.parse_with_lxml(content, **kwargs)
            else:
                raise
        
        return res

    @classmethod
    def parse_with_xmltodict(cls, content, **parse_kwargs):
        return xmltodict.parse(content, **parse_kwargs)

    @classmethod
    def parse_with_lxml(cls, content, **parser_kwargs):
        parser = etree.XMLParser(**parser_kwargs)
        node = etree.fromstring(content, parser)
        return cls._lxml_node2dict(node)

    
    @classmethod
    def unparse(cls, data, pretty=True, indent=Tools.indent_string(4, ''), **unparse_kwargs):
        """ Unparse dict to string, in libraries order and if available:
          - xmltodict,
          - lxml.
        """
        res = to_be_defined
        
        if with_xmltodict:
            kwargs = unparse_kwargs['xmltodict'] if 'xmltodict' in unparse_kwargs else {}
            res = cls.unparse_with_xmltodict(data, pretty=pretty, indent=indent, **kwargs)
        elif with_lxml:
            kwargs = unparse_kwargs['lxml'] if 'lxml' in unparse_kwargs else {}
            res = cls.unparse_with_lxml(data, pretty=pretty, indent=indent, **kwargs)
        else:
            raise TechnicalException(f"No XML library is available in (xmltodict, lxml)")
        
        return res
    
    @classmethod
    def unparse_with_xmltodict(cls, data, pretty=True, indent=Tools.indent_string(4, ''), **unparse_kwargs):
        unparse_kwargs['pretty'] = pretty
        unparse_kwargs['indent'] = indent
        
        return xmltodict.unparse(data, **unparse_kwargs)

    @classmethod
    def unparse_with_lxml(cls, data, pretty=True, indent=Tools.indent_string(4, ''), **unparse_kwargs):
        """ Unparse dict with lxml.
        Warning: this method currently doesn't work since method _lxml_dict2node is not implemented.
        """
        unparse_kwargs['pretty_print'] = pretty
        
        node = cls._lxml_dict2node(data)
        res = etree.tostring(node, **unparse_kwargs)
        
        # By default 'tostring' return bytes whereas we expect a string
        if isinstance(res, bytes):
            res = res.encode('utf-8')
        # TODO: manage indent: with pretty=True, etree.tostring indent with 4 spaces -> replace each 4 space by indent value
        
        return res
    
    @classmethod
    def _lxml_node2dict(cls, node, attributes=True, add_node_name_as_key=True):
        """
        Convert a lxml.etree node tree into a dict.
        """
        result = {}
        if attributes:
            for item in node.attrib.items():
                key, result[key] = item
        
        for element in node.iterchildren():
            # Remove namespace prefix
            key = etree.QName(element).localname
    
            # Process element as tree element if the inner XML contains non-whitespace content
            if element.text and element.text.strip():
                value = element.text
            else:
                value = cls._lxml_node2dict(element, attributes=attributes, add_node_name_as_key=False)
            if key in result:
                if type(result[key]) is list:
                    result[key].append(value)
                else:
                    result[key] = [result[key], value]
            else:
                result[key] = value
        
        # Surround result by node name if requested 
        if add_node_name_as_key:
            name = etree.QName(node).localname
            res = {name:result}
        else:
            res = result
        
        return res

    @classmethod
    def _lxml_dict2node(cls, data, attributes=True):
        """
        Convert a dict into a lxml.etree node tree.
        """
        raise NotImplementedError()






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_multitask.multithreading.reflection.sys import get_current_function_name
from holado_ui.ui.gui.inspectors.gui_inspector import GUIInspector
import abc
from holado_ui.ui.inspectors.tools.ui_inspect_context import UIInspectContext
from holado_core.common.inspectors.tools.inspect_parameters import InspectParameters
from holado_ui_selenium.ui.gui.selenium.criterias.selenium_criteria import SeleniumCriteria
from holado_core.common.criterias.and_criteria import AndCriteria
from holado_ui_selenium.ui.gui.selenium.finders.parent_selenium_finder import ParentSeleniumFinder
from holado_ui_selenium.ui.gui.selenium.finders.children_selenium_finder import ChildrenSeleniumFinder
import re
from holado_ui_selenium.ui.gui.selenium.finders.by_selenium_finder import BySeleniumFinder

logger = logging.getLogger(__name__)



class SeleniumInspector(GUIInspector):
    """ Base class for Selenium inspector.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, module_name):
        super().__init__(module_name)
    
    def get_finder_parent(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        return ParentSeleniumFinder()
    
    def get_finder_children(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        return ChildrenSeleniumFinder()
    
    def get_finder_select(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder for select elements
        """
        return self._get_finder_from_modules(get_current_function_name(), [], [], inspect_context, inspect_parameters)
    
    def get_finder_by_content_text(self, text, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param text Text
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder for element with given text as content
        """
        criteria = SeleniumCriteria(self)
        criteria.content_text = text
        res = BySeleniumFinder(criteria=criteria, description=f"text with content '{text}'")
        
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    def get_finder_by_node_text(self, text, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param text Text
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder for node with given text
        """
        # Path to the node, containing a text node, equals to the text after trimming
        xpath = ".//text()[normalize-space(.) = concat('{}', '')]/..".format(text.replace("'", "', \"'\", '"))
        
        # Add verification, because if only xpath is used hidden texts (ex: not selected select options) are also found
        criteria = SeleniumCriteria(self)
        criteria.content_text = text
        
        res = BySeleniumFinder(xpath=xpath, criteria=criteria, description=f"text '{text}'")
        
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    def get_finder_by_class_name(self, class_name, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Finder by class name.
        @param class_name Class name.
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder.
        """
        criteria = SeleniumCriteria(self)
        criteria.class_name = class_name
        res = BySeleniumFinder(criteria=criteria, description=f"element with class '{class_name}'")
        
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    def get_finder_by_id(self, id_, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param id Id
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return finder for an element with given id
        """
        criteria = SeleniumCriteria(self)
        criteria.id = id_
        res = BySeleniumFinder(criteria=criteria, description=f"element with id '{id_}'")
        
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    def get_finder_by_tag(self, tag_name, element_description=None, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Finder for an element with given tag name
        @param tag_name Tag name
        @param element_description Element description
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder
        """
        if element_description is None:
            element_description = tag_name
        
        criteria = SeleniumCriteria(self)
        criteria.tag_name = tag_name
        res = BySeleniumFinder(criteria=criteria, description=element_description)
        
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    def get_finder_by_tag_and_attribute(self, tag_name, attr_name, attr_value, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Finder by given attribute.
        @param tag_name Tag name
        @param attr_name Attribute name.
        @param attr_value Attribute value.
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder.
        """
        xpath = f"//{tag_name}[@{attr_name}='{attr_value}']"
        return self.get_finder_by_xpath(xpath, f"{tag_name} with attribute '{attr_name}' = '{attr_value}'", inspect_context, inspect_parameters)
    
    def get_finder_by_tag_and_inner_text(self, tag_name, inner_text, element_description=None, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param tag_name Tag name
        @param inner_text Inner text
        @param element_description Element description
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder for an element with given tag and inner text
        """
        if element_description is None:
            element_description = f"{tag_name} '{inner_text}'"
        
        criteria = SeleniumCriteria(self)
        criteria.tag_name = tag_name
        criteria.inner_text = inner_text
        
        res = BySeleniumFinder(criteria=criteria, description=element_description)
        
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    def get_finder_by_attribute(self, attr_name, attr_value, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Finder by given attribute.
        @param attr_name Attribute name.
        @param attr_value Attribute value.
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder.
        """
        xpath = f"//*[@{attr_name}='{attr_value}']"
        return self.get_finder_by_xpath(xpath, f"element with attribute '{attr_name}' = '{attr_value}'", inspect_context, inspect_parameters)
    

    def get_finder_by_xpath(self, xpath, element_description, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Finder by given XPath.
        @param xpath XPath
        @param element_description Element description
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder.
        """
        res = BySeleniumFinder(xpath=xpath, description=element_description)
        
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.resource.persisted_data_manager import PersistedDataManager
import os
from holado_python.standard_library.csv import CsvManager
from holado_core.common.exceptions.technical_exception import TechnicalException

logger = logging.getLogger(__name__)


class AISManager(object):
    
    def __init__(self):
        self.__persisted_data_manager = PersistedDataManager(data_name="AIS data",
                                                             table_name=self.__get_persisted_data_table_name(),
                                                             table_sql_create=self.__get_persisted_data_table_sql_create(),
                                                             db_name="ais")
        self.__possible_mids = []
        
    def initialize(self, resource_manager):
        self.__persisted_data_manager.initialize(resource_manager)
        
        self.__possible_mids = self.__load_mids()
    
    def __load_mids(self):
        """Return possible MID (Maritime Identification Digits) values.
        Resource file MaritimeIdentificationDigits.csv is an export from https://www.itu.int/en/ITU-R/terrestrial/fmd/Pages/mid.aspx
        """
        here = os.path.abspath(os.path.dirname(__file__))
        path = os.path.normpath(os.path.join(here, "MaritimeIdentificationDigits.csv"))
        
        table = CsvManager.table_with_content_of_CSV_file(path, delimiter=',')
        return table.get_column(name='Digit').cells_content
    
    def new_unique_mmsi_number(self):
        # Ensure persistent DB exists
        self.__persisted_data_manager.ensure_persistent_db_exists()
        
        last_mmsi = self.__get_persisted_data_value("unique_mmsi")
        res = self.next_valid_mmsi(last_mmsi)
        if res is None:
            # This case appears when last unique MMSI was last valid MMSI. Get first possible MMSI.
            res = self.next_valid_mmsi(None)
            
        self.__update_or_add_persisted_data("unique_mmsi", res)
        
        return res
    
    def format_mmsi_number_as_string(self, mmsi):
        if not isinstance(mmsi, int):
            raise TechnicalException(f"MMSI is not an integer (obtained type: {type(mmsi)})")
        return f"{mmsi:09d}"

    def first_valid_mmsi(self):
        return int(self.__possible_mids[0]) * 10000
    
    def next_valid_mmsi(self, previous_mmsi=None):
        """Return next valid MMSI.
        
        Notes:
            If previous_mmsi is None, return first valid MMSI number.
            If previous_mmsi is last valid MMSI, return None.
        """
        if previous_mmsi is None:
            # Return smallest possible MMSI
            return self.first_valid_mmsi()
        else:
            for mmsi in range(previous_mmsi+1, int(1e9)):
                if self.is_valid_mmsi(mmsi):
                    return mmsi
            return None
    
    def is_valid_mmsi(self, mmsi):
        """Return if MMSI is valid
        The MMSI uniquely identifies a ship (or a group of ships), or a coast radio station.
        MMSI stands for Maritime Mobile Service Identity.
        Possible values:
        * 97<nnnnnnn>: valid
        * [00|98|99]<MID><nnnn> : valid if MID is valid
        * [0|8]<MID><nnnnn> : valid if MID is valid
        * <MID><nnnnnn> : valid if MID is valid
        * <any other value (incl. > 999999999) : invalid
        @param mmsi: MMSI to validate (type: int, str)
        """
        if mmsi is None:
            return False
        
        if isinstance(mmsi, int):
            mmsi = self.format_mmsi_number_as_string(mmsi)
        if mmsi[0:2] in ['97', '99']:
            return True
        else:
            if mmsi[0:2] in ['00', '98', '99']:
                mid = mmsi[2:5]
            elif mmsi[0:1] in ['0', '8']:
                mid = mmsi[1:4]
            else:
                mid = mmsi[0:3]
            return self.is_valid_mid(mid)
    
    def is_valid_mid(self, mid):
        return mid in self.__possible_mids

    def __next_mid(self, previous_mid):
        """Return next MID in possible MID values"""
        if previous_mid is None:
            return self.__possible_mids[0]
        elif previous_mid in self.__possible_mids:
            index = self.__possible_mids.index(previous_mid)
            if index + 1 < len(self.__possible_mids):
                return self.__possible_mids[index+1]
            else:
                return None
        else:
            raise TechnicalException(f"MID {previous_mid} is not valid")

    # Manage persistent data

    def __get_persisted_data_table_name(self):
        return "ais_data"

    def __get_persisted_data_table_sql_create(self):
        return """CREATE TABLE ais_data (
                key text NOT NULL,
                value text
            )"""

    def __update_or_add_persisted_data(self, key, value):
        self.__persisted_data_manager.update_or_add_persisted_data(persisted_filter_data={'key':key}, data={'value':value})

    def __get_persisted_data_value(self, key):
        persisted_data = self.__persisted_data_manager.get_persisted_data(filter_data={'key':key})
        if persisted_data is not None:
            return int(persisted_data['value'])
        else:
            return None
        





# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.scenario.step_tools import StepTools
from holado.common.context.session_context import SessionContext
from holado_test.behave.behave import *  # @UnusedWildImport
from holado_core.common.tables.table_manager import TableManager
from holado_value.common.tables.comparators.table_2_value_table_comparator import Table2ValueTable_Comparator
from holado_core.common.tools.comparators.comparator import CompareOperator
from holado_core.common.tables.table import Table
from holado_core.common.tools.comparators.object_comparator import ObjectComparator
from holado_core.common.tables.table_with_header import TableWithHeader
from holado_value.common.tables.value_table_with_header import ValueTableWithHeader
from holado_value.common.tables.comparators.table_2_value_table_with_header_comparator import Table2ValueTable_WithHeaderComparator
from holado_core.common.tables.comparators.table_with_header_comparator import TableWithHeaderComparator
from holado_value.common.tables.value_table import ValueTable
from holado_core.common.tables.comparators.table_comparator import TableComparator
import logging
from holado_core.common.exceptions.verify_exception import VerifyException
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_python.common.tools.comparators.string_comparator import StringComparator
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_text_interpreter():
    return __get_scenario_context().get_text_interpreter()

def __get_unique_value_manager():
    return SessionContext.instance().unique_value_manager

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()




# @Then(r'\'{object_string}\' is {type_name:S}')
# def step_impl(context, object_string, type_name):
#     obj = __get_text_interpreter().interpret(object_string)
# 
#     # Get class type
#     class_ = Typing.get_class_for_type_name(type_name)
#     
#     if isinstance(obj, class_):
#         if isinstance(obj, basestring):
#             # Verify scenario table structure
#             if len(context.table.headings) != 1 or len(context.table.rows) != 1:
#                 raise FunctionalException("For a string, scenario table must be a one column table with header and only one row")
#             
#             # Verify strings
#             cell = ValueTableCell(context.table.rows[0].cells[0])
#             expected = cell.value
#             if obj != expected:
#                 raise FunctionalException("String value is wrong (expected: '{}' ; obtained: '{}')".format(expected, obj))
#         else:
#             raise NotImplementedError(u"Check content of type '{}'".format(Typing.get_object_class_fullname(obj)))
#     else:
#         raise FunctionalException("Object is of type '{}' (expected type: '{}')".format(Typing.get_object_class_fullname(obj), type_name))



@Then(r"(?P<value_1>{Any}) (?P<operator>==|!=|<|<=|>|>=) (?P<value_2>(?:(?! \(as ).)+)")
def step_impl(context, value_1, operator, value_2):  # @DuplicatedSignature
    val1 = StepTools.evaluate_scenario_parameter(value_1)
    val2 = StepTools.evaluate_scenario_parameter(value_2)
    operator = CompareOperator(operator)
    
    if isinstance(val1, Table) and isinstance(val2, Table):
        if isinstance(val1, TableWithHeader) and isinstance(val2, TableWithHeader):
            if isinstance(val2, ValueTableWithHeader):
                comparator = Table2ValueTable_WithHeaderComparator()
            else:
                comparator = TableWithHeaderComparator()
        else:
            if isinstance(val2, ValueTable):
                comparator = Table2ValueTable_Comparator()
            else:
                comparator = TableComparator()
    elif isinstance(val1, str) and isinstance(val2, str):
        comparator = StringComparator()
    else:
        comparator = ObjectComparator()
        
    comparator.compare(val1, operator, val2)



@Then(r"(?P<value>{Str}) is list")
def step_impl(context, value):  # @DuplicatedSignature
    val = StepTools.evaluate_scenario_parameter(value)
    if not isinstance(val, list):
        raise VerifyException(f"Value [{val}] (from expression [{value}]) is not a list (type: {Typing.get_object_class_fullname(val)})")
    
    if hasattr(context, "table") and context.table is not None:
        table = BehaveStepTools.convert_step_table_2_value_table(context.table)
        obtained = TableManager.convert_list_2_column_table(val)
        comparator = Table2ValueTable_Comparator()
        comparator.equals(obtained, table)

@Then(r"(?P<value>{Str}) is empty list")
def step_impl(context, value):  # @DuplicatedSignature
    val = StepTools.evaluate_scenario_parameter(value)
    if not isinstance(val, list):
        raise VerifyException(f"Value [{val}] (from expression [{value}]) is not a list (type: {Typing.get_object_class_fullname(val)})")
    
    if len(val) > 0:
        raise VerifyException(f"List is not empty, it contains {len(val)} elements: [{val}] (from expression [{value}])")


# @Then(r"dictionary (?P<value>{Str}) is")
# def step_impl(context, value):  # @DuplicatedSignature
#     val = StepTools.evaluate_scenario_parameter(value)
#     if not isinstance(val, dict):
#         raise VerifyException(f"Value [{val}] (from expression [{value}]) is not a dictionary (type: {Typing.get_object_class_fullname(val)})")
#
#     if hasattr(context, "table") and context.table is not None:
#         table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
#         ValueTableManager.verify_table_is_x_table(table, "Key", "Value")
#         table.sort(names=['Key'])
#
#         obtained = TableManager.convert_dict_2_name_value_table(val)
#         obtained.sort(names=['Name'])
#
#         comparator = Table2ValueTable_Comparator()
#         comparator.equals(obtained, table)

@Then(r"(?P<value>{Str}) is text")
def step_impl(context, value):  # @DuplicatedSignature
    obtained = StepTools.evaluate_scenario_parameter(value)
    if not isinstance(obtained, str):
        raise VerifyException(f"Value [{obtained}] (from expression [{value}]) is not a text (type: {Typing.get_object_class_fullname(obtained)})")
    
    # Note: in current implementation of BehaveStepTools.get_step_multiline_text, multiline text is stripped,
    #       thus value must be also stripped
    obtained = obtained.strip()
    
    text = BehaveStepTools.get_step_multiline_text(context)
    if text is not None:
        comparator = StringComparator()
        comparator.equals(obtained, text)







#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import threading
from holado.holado_config import Config
import logging
from holado.common.context.session_context import SessionContext
import time
from holado_multitask.multitasking.multitask_manager import MultitaskManager
from holado_python.standard_library.typing import Typing
from holado.common.handlers.undefined import undefined_argument, default_context
from holado_core.common.exceptions.technical_exception import TimeoutTechnicalException
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_python.common.tools.datetime import DateTime, FORMAT_DATETIME_ISO
from holado.common.handlers.object import DeleteableObject

logger = logging.getLogger(__name__)


class Thread(threading.Thread, DeleteableObject):
    """ Helper class managing a thread."""
    
    def __init__(self, name, default_wait_timeout=None, register_thread=True, register_scope=default_context, delay_before_run_sec=None, delay_between_run_sec=None):
        threading.Thread.__init__(self)
        DeleteableObject.__init__(self, name)
        
        if default_wait_timeout is None:
            default_wait_timeout = Config.timeout_seconds * 1000
            
        # Note: if SessionContext doesn't have an instance (like during session context reset), use name+now as unique name
        self.__unique_name = SessionContext.instance().multitask_manager.prepare_thread(name) if SessionContext.has_instance() else f"{name}_{DateTime.datetime_2_str(DateTime.now(), FORMAT_DATETIME_ISO)}"
        self.__default_wait_timeout = default_wait_timeout
        self.__delay_before_run_sec = delay_before_run_sec
        self.__delay_between_run_sec = delay_between_run_sec

        # Put thread as daemon
        self.daemon = True
        
        # Manage thread status
        self._is_lock = threading.Lock()
        self._is_idle = threading.Event()
        self._is_idle.set()
        
        # Register thread
        # Note: Do not register thread if session context doesn't have an instance (like during session context reset)
        self.__is_registered = False
        self.__register_scope = None
        if register_thread and SessionContext.has_instance():
            # Registered name has to be unique
            SessionContext.instance().threads_manager.register_thread(self.__unique_name, self, scope=register_scope)
            self.__is_registered = True
            self.__register_scope = register_scope
            self.on_delete_call_func(self.__unregister_thread)
        
    @property
    def unique_name(self):
        return self.__unique_name

    @property
    def default_wait_timeout(self):
        return self.__default_wait_timeout
        
    @default_wait_timeout.setter
    def default_wait_timeout(self, timeout):
        self.__default_wait_timeout = timeout
        
    @property
    def delay_before_run_seconds(self):
        return self.__delay_before_run_sec
        
    @property
    def delay_between_run_seconds(self):
        return self.__delay_between_run_sec
        
    def _set_ident(self):
        super()._set_ident()
        if not MultitaskManager.has_thread_native_id() and SessionContext.has_instance():
            SessionContext.instance().multitask_manager.set_thread_uid(self.__unique_name, MultitaskManager.get_thread_uid(thread=self))
        
    if MultitaskManager.has_thread_native_id():
        def _set_native_id(self):
            super()._set_native_id()
            if SessionContext.has_instance():
                SessionContext.instance().multitask_manager.set_thread_uid(self.__unique_name, MultitaskManager.get_thread_uid(thread=self))
        
    def join(self, timeout=undefined_argument, raise_if_still_alive=True):
        """
        See Thread.join documentation for default behavior. 
        
        This implementation returns if join succeeded, ie if thread is still alive after join.
        
        This implementation also tries to avoid deadlocks by special and default timeout value undefined_argument.
        With any other timeout value, including None, default behavior is happening.
        If timeout argument is undefined_argument, it means that it should stop by itself without deadlock.
        In this case, a timeout value Config.join_timeout_seconds is used.
        
        If argument raise_if_still_alive is True (default) after join and if thread is still alive after join, an exception is raised.
        If timeout is undefined_argument, a TimeoutTechnicalException is raised, else a FunctionalException.
        """
        timeout_str = "long and finite" if timeout is undefined_argument else f"{timeout}s" if timeout else "none"
        logger.debug(f"Joining thread '{self.name}' ; timeout={timeout_str} ...")
        
        real_timeout = timeout if timeout is not undefined_argument else Config.join_timeout_seconds
        super().join(real_timeout)
        
        logger.debug(f"End of joining thread '{self.name}' ; timeout={timeout_str} ; alive:{self.is_alive()}")
        
        if self.is_alive():
            if raise_if_still_alive:
                if timeout is undefined_argument:
                    raise TimeoutTechnicalException(f"Thread '{self.name}' is still alive after join (timeout: {real_timeout} s)")
                else:
                    raise FunctionalException(f"Thread '{self.name}' is still alive after join (timeout: {real_timeout} s)")
            else:
                logger.error(f"Thread '{self.name}' is still alive after join (timeout: {real_timeout} s)")
        
        return self.is_alive()
    
    def wait_before_run(self, do_log_exception=True):
        if self.delay_before_run_seconds is not None:
            try:
                time.sleep(self.delay_before_run_seconds)
            except Exception:
                if do_log_exception:
                    logger.exception(f"[Thread '{self.name}'] Exception while waiting delay before run (delay: {self.delay_before_run_seconds} ; type: {Typing.get_object_class_fullname(self.delay_before_run_seconds)})")

    def wait_between_run(self, do_log_exception=True):
        if self.delay_between_run_seconds is not None:
            try:
                time.sleep(self.delay_between_run_seconds)
            except Exception:
                if do_log_exception:
                    logger.exception(f"[Thread '{self.name}'] Exception while waiting delay between run (delay: {self.delay_between_run_seconds} ; type: {Typing.get_object_class_fullname(self.delay_between_run_seconds)})")

    def wait_is_idle(self, timeout=None):
        """ Wait event self._is_idle"""
        self._wait(self.__is_idle, "is idle", timeout=timeout)

    def _wait(self, event, event_details, timeout=None):
        """ Wait given threading.Event"""
        is_set = False
        wait_round_counter = 0
        round_wait_seconds = 1
        if timeout is None:
            timeout = self.__default_wait_timeout
        
        logger.debug("[{}] Waiting {}...".format(self.name, event_details))
        while not is_set:
            wait_round_counter += 1
            is_set = event.wait(round_wait_seconds)
            
            if not is_set:
                if wait_round_counter * round_wait_seconds * 1000 > timeout:
                    raise TimeoutError("[{}] Timeout when waiting {}".format(self.name, event_details))
                elif wait_round_counter % 10 == 0:
                    logger.debug("[{}] Still waiting {} (round {})".format(self.name, event_details, wait_round_counter))
        logger.debug("[{}] Finished waiting {}...".format(self.name, event_details))
    
    def __unregister_thread(self):
        if self.__is_registered:
            if SessionContext.instance().threads_manager.is_thread_registered(self.__unique_name, scope=self.__register_scope):
                SessionContext.instance().threads_manager.unregister_thread(self.__unique_name, scope=self.__register_scope)
            self.__is_registered = False


class InterruptableThread(Thread):
    """ Helper class managing a thread interruptable."""
    
    def __init__(self, name, default_wait_timeout=None, register_thread=True, register_scope=default_context, delay_before_run_sec=None, delay_between_run_sec=None):
        super().__init__(name, default_wait_timeout=default_wait_timeout, register_thread=register_thread, register_scope=register_scope, delay_before_run_sec=delay_before_run_sec, delay_between_run_sec=delay_between_run_sec)

        # Manage interrupt
        self.__interrupt_lock = threading.Lock()
        self.__interrupt_event = threading.Event()
    
    @property    
    def is_interruptable(self):
        return True
    
    @property
    def is_interrupted(self):
        with self.__interrupt_lock:
            return self.__interrupt_event.is_set()

    def interrupt(self):
        if not self.is_interrupted:
            logger.debug("Interrupt thread '{}'".format(self.name))
            with self.__interrupt_lock:
                self.__interrupt_event.set()
    
    def _raise_if_interrupted(self):
        if self.is_interrupted:
            raise InterruptedError




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import json
from holado_core.common.resource.persisted_data_manager import PersistedDataManager
from holado_python.standard_library.typing import Typing
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools
from holado.common.handlers.undefined import undefined_value, undefined_argument,\
    any_value
from holado_python.common.tools.datetime import DateTime
from holado_db.tools.db.query.base.query_builder import DBCompareOperator
from holado.common.handlers.enums import AutoNumber

logger = logging.getLogger(__name__)


class MethodCallStatus(AutoNumber):
    Ready = ()
    Completed = ()
    Failed = ()
    Expired = ()
    

class PersistedMethodToCallManager(PersistedDataManager):
    def __init__(self, scope_name, delete_on_success=True, delete_on_fail=False, delete_on_success_after_fail=False, expiration_try_counter=10, data_name="method", table_name="method_to_call", db_name="default"):
        """ Constructor
        @param scope_name: Scope of this manager
        
        Note: if scope_name is any_value, it is possible to call all methods persisted with any scope, but it is not possible to add new methods.
        """
        super().__init__(data_name=data_name, table_name=table_name, 
                         table_sql_create=self._get_default_table_sql_create(table_name), 
                         db_name=db_name)
        self.__scope_name = scope_name
        self.__delete_on_success = delete_on_success
        self.__delete_on_fail = delete_on_fail
        self.__delete_on_success_after_fail = delete_on_fail
        self.__expiration_try_counter = expiration_try_counter
    
    def initialize(self, resource_manager, expression_evaluator):
        super().initialize(resource_manager)
        self.__expression_evaluator = expression_evaluator
    
    def _get_default_table_sql_create(self, table_name):
        return f"""CREATE TABLE {table_name} (
                id INTEGER PRIMARY KEY,
                scope_name text NOT NULL,
                function_qualname text NOT NULL,
                self_getter TEXT,
                args TEXT,
                kwargs TEXT,
                use text NOT NULL,
                use_index integer NOT NULL,
                created_at TEXT,
                changed_at TEXT,
                status TEXT,
                try_counter INTEGER,
                error TEXT,
                do_delete_on_success INTEGER,
                do_delete_on_fail INTEGER,
                do_delete_on_success_after_fail INTEGER
            )"""
    
    def add_function_to_call(self, function_qualname, args_list=None, kwargs_dict=None, use="default", use_index=0, add_if_exists=False, delete_on_success=undefined_argument, delete_on_fail=undefined_argument, delete_on_success_after_fail=undefined_argument):
        """Add a function to call.
        @param function_qualname: Qualified name of function
        @param args_list: List of function args (default: None)
        @param kwargs_dict: Dict of function kwargs (default: None)
        @param use: Define persistent usage. It usually corresponds to a specific scope.
        @param use_index: use index, useable to order the functions to call. By default all are 0. If set to None, it is automatically set to max(use_index)+1.
        @param delete_on_success: If True, delete function after its success
        @param delete_on_fail: If True, delete function after its fail
        @param delete_on_success_after_fail: If True, delete function after its success even if it has previously failed
        """
        self.add_method_to_call(function_qualname, None, args_list, kwargs_dict, use, use_index, add_if_exists, delete_on_success, delete_on_fail, delete_on_success_after_fail)
    
    def add_method_to_call(self, function_qualname, self_getter_eval_str, args_list=None, kwargs_dict=None, use="default", use_index=0, add_if_exists=False, delete_on_success=undefined_argument, delete_on_fail=undefined_argument, delete_on_success_after_fail=undefined_argument):
        """Add a method to call.
        @param function_qualname: Qualified name of function
        @param self_getter_eval_str: String to eval in order to get the self instance to use when calling method 
        @param args_list: List of function args (default: None)
        @param kwargs_dict: Dict of function kwargs (default: None)
        @param use: Define persistent usage. It usually corresponds to a specific scope.
        @param use_index: Use index, useable to order the functions to call. By default all are 0. If set to None, it is automatically set to max(use_index)+1.
        @param add_if_exists: Add if method already exists (default: False)
        @param delete_on_success: If True, delete method after its success
        @param delete_on_fail: If True, delete method after its fail
        @param delete_on_success_after_fail: If True, delete method after its success even if it has previously failed
        """
        if self.__scope_name is any_value:
            raise TechnicalException("To add a method to call, the scope name must be specified")
        
        # Define persisted data that can define if method is already persisted
        data = {
            'scope_name': self.__scope_name,
            'function_qualname': function_qualname,
            'self_getter': self_getter_eval_str,
            'use': use,
            }
        if args_list is not None:
            data['args'] = json.dumps(args_list)
        if kwargs_dict is not None:
            data['kwargs'] = json.dumps(kwargs_dict)
        
        # Return if data is already persisted and add_if_exists==True
        filter_compare_data = [('status', DBCompareOperator.In, [MethodCallStatus.Ready.name, MethodCallStatus.Failed.name])]
        if not add_if_exists and self.has_persisted_data(filter_data=data, filter_compare_data=filter_compare_data):
            return
        
        # Persist data
        if use_index is None:
            use_index = self.__get_use_next_index(use)
        now_str = DateTime.datetime_2_str(DateTime.utcnow())
        data.update({
            'use_index': use_index,
            'created_at': now_str,
            'changed_at': now_str,
            'status': MethodCallStatus.Ready.name,
            'try_counter': 0,
            'do_delete_on_success': self.__get_db_on_delete_after_when_add_method(delete_on_success, self.__delete_on_success),
            'do_delete_on_fail': self.__get_db_on_delete_after_when_add_method(delete_on_fail, self.__delete_on_fail),
            'do_delete_on_success_after_fail': self.__get_db_on_delete_after_when_add_method(delete_on_success_after_fail, self.__delete_on_success_after_fail)
            })
        
        self.add_persisted_data(data)
    
    def get_number_of_functions_and_method_to_call(self, use=undefined_argument):
        methods_data = self.__get_functions_and_methods_to_call(use=use)
        return len(methods_data)
    
    def __get_db_on_delete_after_when_add_method(self, method_delete, default_delete):
        if method_delete is not undefined_argument:
            if method_delete is undefined_value:
                return -1
            else:
                return method_delete
        else:
            if default_delete is undefined_value:
                return -1
            else:
                return default_delete
        
    def __get_do_delete_after_call(self, db_delete, default_delete):
        if db_delete != -1:
            return db_delete
        else:
            if default_delete is undefined_value:
                return False
            else:
                return default_delete
        
    def __get_use_next_index(self, use):
        datas = self.get_persisted_datas({'scope_name':self.__scope_name, 'use':use})
        if datas:
            return max(map(lambda x:x['use_index'], datas)) + 1
        else:
            return 0
    
    def __get_functions_and_methods_to_call(self, use=undefined_argument, use_index=None):
        filter_data = {}
        if self.__scope_name is not any_value:
            filter_data['scope_name'] = self.__scope_name
        if use is not undefined_argument:
            filter_data['use'] = use
            if use_index is not None:
                filter_data['use_index'] = use_index
        filter_compare_data = [('status', DBCompareOperator.In, [MethodCallStatus.Ready.name, MethodCallStatus.Failed.name])]
        return self.get_persisted_datas(filter_data, filter_compare_data=filter_compare_data)
    
    def call_functions_and_methods(self, use="default", use_index=None):
        """Call methods of given use
        @param use: Define persistent usage. It usually corresponds to a specific scope.
        @param use_index: If defined, call only functions and methods of given index.
        """
        # Get functions and methods to call
        methods_data = self.__get_functions_and_methods_to_call(use=use)
        
        # Call methods
        if methods_data:
            for meth_index, meth_data in enumerate(methods_data):
                status, do_delete, error = None, None, ''
                try:
                    self._call_function_or_method(meth_data)
                except Exception as exc:
                    error = Tools.represent_exception(exc, indent=8)
                    msg_list = [f"Error while calling following method (use: '{use}' ; use index: {use_index} ; method index: {meth_index}):"]
                    msg_list.append(Tools.represent_object(meth_data, 8))
                    msg_list.append("    Error:")
                    msg_list.append(error)
                    msg_list.append("  => Continue to process persisted methods")
                    msg_list.append("     WARNING: this method is removed from persisted methods to avoid recursive and blocking errors")
                    logger.error("\n".join(msg_list))
                    status = MethodCallStatus.Failed
                    do_delete = self.__get_do_delete_after_call(meth_data['do_delete_on_fail'], self.__delete_on_fail)
                else:
                    status = MethodCallStatus.Completed
                    if meth_data['try_counter'] == 0:
                        do_delete = self.__get_do_delete_after_call(meth_data['do_delete_on_success'], self.__delete_on_success)
                    else:
                        do_delete = self.__get_do_delete_after_call(meth_data['do_delete_on_success_after_fail'], self.__delete_on_success_after_fail)
                
                self.__update_persisted_data_after_call(do_delete, meth_data, status, error)
                
    def __update_persisted_data_after_call(self, do_delete, meth_data, status, error):
        if do_delete:
            self.__delete_function_or_method(meth_data)
        else:
            self.__update_function_or_method_status(meth_data, status, error)
    
    def __delete_function_or_method(self, function_or_method_data):
        filter_data = {'id':function_or_method_data['id']}
        self.delete_persisted_data(filter_data)
    
    def __update_function_or_method_status(self, function_or_method_data, status, error):
        filter_data = {'id':function_or_method_data['id']}
        try_counter = function_or_method_data['try_counter'] + 1
        if try_counter >= self.__expiration_try_counter and status == MethodCallStatus.Failed:
            status = MethodCallStatus.Expired
        
        data = {
            'status': status.name,
            'try_counter': try_counter,
            'error': error,
            'changed_at': DateTime.datetime_2_str(DateTime.utcnow()),
            }
        self.update_persisted_data(data, filter_data=filter_data)
    
    def _call_function_or_method(self, function_or_method_data):
        _, func = self.__expression_evaluator.evaluate_python_expression(function_or_method_data['function_qualname'])
        if not Typing.is_function(func):
            raise TechnicalException(f"Failed to evaluate python expression '{function_or_method_data['function_qualname']}' as a function (obtained: {func} [type: {Typing.get_object_class_fullname(func)}] ; function data: {function_or_method_data})")
        
        func_self = None
        if function_or_method_data['self_getter'] is not None:
            _, func_self = self.__expression_evaluator.evaluate_python_expression(function_or_method_data['self_getter'])
        
        if function_or_method_data['args'] is not None:
            args = json.loads(function_or_method_data['args'])
        else:
            args = []
        if function_or_method_data['kwargs'] is not None:
            kwargs = json.loads(function_or_method_data['kwargs'])
        else:
            kwargs = {}
        
        if func_self is not None:
            func(func_self, *args, **kwargs)
        else:
            func(*args, **kwargs)
        
    



# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

# This file is a helper to build scripts using HolAdo solution, with CSV files as input.
# Usually, it is copied and adapted for specific usage.


import logging
import os
from holado_helper.script.initialize_script import get_logging_argument_parser
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)

class Script(object):
    
    def __init__(self, with_args=True):
        self.__with_args = with_args
        
        self.__args = None
        self.__cwd = None
        self.__user_work_dir=None
    
    @property
    def args(self):
        return self.__args
    
    @property
    def cwd(self):
        """ Current working directory
        A script can change its working directory by calling 'change_working_dir' method.
        """
        if self.__cwd is not None:
            return self.__cwd
        else:
            return os.getcwd()
    
    @property
    def user_working_dir(self):
        """ User working directory
        Although the user working directory is persisted.
        """
        if self.__user_work_dir is not None:
            return self.__user_work_dir
        else:
            return os.getcwd()
    
    @user_working_dir.setter
    def user_working_dir(self, wd):
        self.__user_work_dir = wd
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Changed user working dir to '{self.__user_work_dir}'")
    
    def parse_args(self):
        arg_parser = self.new_argument_parser()
        self.__args = arg_parser.parse_args()
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Parsed args: {self.__args}")
    
    def new_argument_parser(self):
        raise NotImplementedError
    
    def _new_argument_parser(self, description=None, additional_arguments=None, add_help=True):
        parser = get_logging_argument_parser(description=description, additional_arguments=additional_arguments, add_help=add_help)
        return parser
    
    def change_working_dir(self, work_dir_path):
        from holado_helper.script.initialize_script import change_working_dir as init_cd    # @UnresolvedImport
        
        # If user working dir is not yet defined, save current working dir as user working dir
        if self.__user_work_dir is None:
            self.user_working_dir = os.getcwd()
        
        # Change working dir
        new_cwd = init_cd(work_dir_path)
        self.__cwd = new_cwd
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Changed script working dir to '{self.__cwd}'")
    
    def initialize(self):
        if self.args is not None and hasattr(self.args, 'log_level') and hasattr(self.args, 'log_in_file'):
            from holado_helper.script.initialize_script import change_logging_config # @UnresolvedImport
            change_logging_config(log_level=self.args.log_level, log_in_file=self.args.log_in_file)
    
    def run(self):
        self.parse_args()
        self.initialize()
        self.run_script()
        
    def run_script(self):
        raise NotImplementedError
    
    def _build_path(self, filename, cwd):
        if os.path.isabs(filename):
            return filename
        else:
            return os.path.join(cwd, filename)






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_rest.api.rest.rest_client import RestClient
import socket
import urllib.parse
import re
from holado_rest.api.rest.rest_manager import RestManager
from holado.common.handlers.undefined import undefined_argument, undefined_value
import os
from holado_core.common.tools.converters.converter import Converter
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.handlers.wait import WaitFuncResult

logger = logging.getLogger(__name__)


class TestServerClient(RestClient):

    @classmethod
    def new_client(cls, use_localhost=undefined_argument, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = None
        if 'url' not in kwargs:
            if use_localhost is undefined_argument:
                env_use = os.getenv("HOLADO_USE_LOCALHOST", False)
                use_localhost = Converter.is_boolean(env_use) and Converter.to_boolean(env_use)
            
            url = os.getenv("HOLADO_TEST_SERVER_URL", undefined_value)
            if url is undefined_value:
                scheme = kwargs.get('scheme', undefined_value)
                if scheme is undefined_value:
                    scheme = os.getenv("HOLADO_TEST_SERVER_SCHEME", "http")
                host = kwargs.get('host', undefined_value)
                if host is undefined_value:
                    host = "localhost" if use_localhost else os.getenv("HOLADO_TEST_SERVER_HOST", "holado_test_server")
                port = kwargs.get('port', undefined_value)
                if port is undefined_value:
                    port = os.getenv("HOLADO_TEST_SERVER_PORT", 51232)
                
                if port is None:
                    url = f"{scheme}://{host}"
                else:
                    url = f"{scheme}://{host}:{port}"
            kwargs['url'] = url
        
        manager = RestManager(default_client_class=cls)
        res = manager.new_client(**kwargs)
        
        return res
    
    
    def __init__(self, name, url, headers=None):
        super().__init__(name, url, headers)
        
        self.__is_available = None
    
    @property
    def is_available(self):
        if self.__is_available is None:
            self.__is_available = self.ping()
            logger.info(f"Test server is {'not ' if not self.__is_available else ''}available")
        return self.__is_available
    
    # Monitoring
    
    def ping(self):
        url_parsed = urllib.parse.urlparse(self.url)
        netloc_pattern = r"(?P<host>.*?)(?::(?P<port>\d+))?$"
        m = re.match(netloc_pattern, url_parsed.netloc)
        host = m.group('host')
        if m.group('port') is not None:
            port = Converter.to_integer(m.group('port'))
        elif url_parsed.scheme == "http":
            port = 80
        elif url_parsed.scheme == "https":
            port = 443
        else:
            raise TechnicalException(f"Unable to define port used by test-server (URL: '{self.url}')")
        
        sock = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
        sock.settimeout(1)
        try:
            sock.connect((host,port))
        except Exception as exc:
            logger.debug(f"Ping of test server failed for ({host}, {port}): {str(exc)}")
            return False
        else:
            sock.close()
            logger.debug(f"Ping of test server succeeded for ({host}, {port})")
            return True
        
    def is_healthy(self):
        try:
            response = self.get(f"health")
            return response.status_code == 200
        except:
            return False
    
    def wait_is_healthy(self, timeout_sec=300, do_raise_on_timeout=True):
        logger.info(f"Waiting for test-server ({self.url}) to be healthy...")
        wait_context = WaitFuncResult(f"wait test-server is healthy", self.is_healthy)
        wait_context.with_timeout(timeout_sec) \
                    .with_raise_on_timeout(do_raise_on_timeout) \
                    .with_process_in_thread(False) \
                    .redo_until(True)
        try:
            wait_context.execute()
        except Exception as exc:
            logger.error(f"Error while waiting for test-server ({self.url}) to be healthy: {str(exc)}")
            raise
        else:
            logger.info(f"test-server ({self.url}) is healthy")
    
    
    # Manage campaigns actions
    
    def update_stored_campaigns(self):
        response = self.put(f"campaign/update")
        return self.response_result(response, status_ok=[200, 204])
    
    def get_scenario_history(self, scenario_name=None, size=None):
        data = {}
        if scenario_name is not None:
            data['scenario_name'] = scenario_name
        if size is not None:
            data['size'] = size
            
        if data:
            response = self.get(f"campaign/scenario/history", json=data)
        else:
            response = self.get(f"campaign/scenario/history", json=data)
        
        return self.response_result(response, status_ok=[200, 204])
    
    
    
    
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.exceptions.technical_exception import TechnicalException
import logging
from holado_core.common.tools.tools import Tools
import json
from holado_core.common.tools.converters.converter import Converter
from holado_json.ipc.json_converter import JsonConverter
from holado.common.handlers.object import Object
from holado.holado_config import Config
from holado.common.handlers.undefined import to_be_defined, default

logger = logging.getLogger(__name__)

try:
    import requests
    from requests.auth import HTTPBasicAuth
    with_requests = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"RestClient is not available. Initialization failed on error: {exc}")
    with_requests = False


class RestClient(Object):
    """
    REST client
    """
    
    default_result_on_statuses = {
        # 204 == "No Content" => result is None
        204: None
        }
    
    @classmethod
    def is_available(cls):
        return with_requests
    
    def __init__(self, name, url, headers=None):
        super().__init__(name)
        self.__url = url.rstrip('/')
        self.__headers = headers if headers is not None else {}
        self.__kwargs = {}
            
        self.__request_log_level = Config.log_level_rest_request
    
    @property
    def url(self):
        return self.__url
    
    @property
    def request_log_level(self):
        return self.__request_log_level
    
    @request_log_level.setter
    def request_log_level(self, log_level):
        self.__request_log_level = log_level
    
    def authenticate_by_user(self, user, pwd):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Authenticate with user ({user},{pwd})")
        self.__kwargs['auth'] = HTTPBasicAuth(user, pwd)
    
    def authenticate_by_token(self, token_type, access_token):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Authenticate with token type '{token_type}' and access token '{access_token}'")
        self.__headers['Authorization'] = f'{token_type} {access_token}'
    
    def with_ssl(self, ssl_kwargs):
        unexpected_keys = set(ssl_kwargs.keys()).difference(['verify', 'cert'])
        if len(unexpected_keys) > 0:
            raise TechnicalException(f"Unmanaged SSL arguments: {unexpected_keys}")
        
        self.__kwargs.update(ssl_kwargs)
    
    def add_parameters_to_path(self, path, parameters:dict):
        res = path
        if parameters:
            res += "&" if '?' in res else "?"
            res += "&".join([f"{k}={v}" for k,v in parameters.items()])
        return res
    
    def response_result(self, response, status_ok=200, result_on_statuses=default_result_on_statuses):
        """Return the result of the request associated to given response
        @param response Response of the request
        @param status_ok The status code, or a list of status codes, for which the request is considered in success
        @param result_on_statuses A dictionary with defined result associated to a status
        Notes:
          - If response status is not status_ok (or in status_ok if it is a list), an exception is raised telling request has failed
          - result_on_statuses is by default RestClient.default_result_on_statuses (={204: None})
          - When holado.common.handlers.undefined.default is in result_on_statuses as status, the associated result is returned in any other case and no failing exception is raised
        """
        res = to_be_defined
        is_res_default = False
        is_ok = isinstance(status_ok, list) and response.status_code in status_ok or response.status_code == status_ok
        
        # Find a result associated to response status
        if result_on_statuses is not None:
            for status, status_result in result_on_statuses.items():
                if response.status_code == status:
                    res = status_result
                    break
                elif status is default and not is_ok:
                    res = status_result
                    is_res_default = True
                    break
                    
        
        # If result is not defined, get it from response
        if res is to_be_defined:
            if "json" in response.headers['Content-Type'] and len(response.content) > 0:
                res = response.json()
            elif response.headers['Content-Type'].startswith('text'):
                res = response.text
            else:
                res = response.content
        
        # Verify status
        if not is_ok and not is_res_default:
            raise FunctionalException(f"[{self.name}] Request failed with status {response.status_code} (expected success status: {status_ok}) on error: {res}")
        
        return res
    
    def request(self, method, path, path_parameters=None, **kwargs):
        path = self.add_parameters_to_path(path, path_parameters)
        url = self.__build_url(path=path)
        r_kwargs = self.__build_kwargs(**kwargs)
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Calling {method} on url '{url}' with arguments {r_kwargs}")
        try:
            res = requests.request(method, url, **r_kwargs)
        except Exception as exc:
            raise TechnicalException(f"[{self.name}] Failed to process {method} '{url}' with arguments {r_kwargs}") from exc
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[{self.name}] {method} on url '{url}' => {Tools.represent_object(res, 4)}", msg_size_limit=None)
        elif Tools.do_log(logger, self.request_log_level):
            logger.log(self.request_log_level, f"[{self.name}] {method} on url '{url}' => {Tools.represent_object(res, 4, max_level=1)}")
        return res
        
    def delete(self, path, data=None, path_parameters=None, **kwargs):
        return self.request("DELETE", path, path_parameters=path_parameters, data=data, **kwargs)
        
    def get(self, path, data=None, path_parameters=None, **kwargs):
        return self.request("GET", path, path_parameters=path_parameters, data=data, **kwargs)
        
    def patch(self, path, data=None, path_parameters=None, **kwargs):
        return self.request("PATCH", path, path_parameters=path_parameters, data=data, **kwargs)
        
    def post(self, path, data=None, json=None, path_parameters=None, **kwargs):
        return self.request("POST", path, path_parameters=path_parameters, data=data, json=json, **kwargs)
        
    def put(self, path, data=None, path_parameters=None, **kwargs):
        return self.request("PUT", path, path_parameters=path_parameters, data=data, **kwargs)
    
    def __build_url(self, path):
        return f"{self.__url}/{path.lstrip('/')}" 
        
    def __build_kwargs(self, **request_kwargs):
        res = dict(self.__kwargs)
        if self.__headers:
            res['headers'] = self.__headers
        if request_kwargs:
            # Add all kwargs except data
            data = request_kwargs.pop('data', None)
            for key in list(request_kwargs.keys()):
                if key in res and isinstance(res[key], dict):
                    res[key].update(request_kwargs.pop(key))
            res.update(request_kwargs)
            
            # Ensure data is in right format
            if data is not None:
                # Convert data to json object for types not supported by requests.request
                if not (isinstance(data, str) or Converter.is_dict(data) or Converter.is_list(data) or isinstance(data, bytes) or Converter.is_file_like(data)):
                    converter = JsonConverter()
                    data = converter.to_json(data)
                
                # Convert data to string for some content types
                if not isinstance(data, str):
                    content_type = res['headers']['Content-Type'] if 'headers' in res and 'Content-Type' in res['headers'] else None
                    if content_type is not None and 'application/json' in content_type:
                        data = json.dumps(data)
                res['data'] = data
            
        return res 
        
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import copy
from holado_core.common.finders.tools.find_updater import FindUpdater
from holado_core.common.finders.tools.enums import FindType
from holado_core.common.handlers.element_holder import ElementHolder
from holado_core.common.finders.tools.find_builder import FindBuilder


class FindContext(object):
    """Base find context
    In usual use cases, the appropriate sub-class should be instantiated.
    """
    
    __instance_default = None
    
    def __init__(self, find_type=None, find_context=None):
        if find_type is not None:
            self.__find_type = find_type
        elif find_context is not None:
            self.__find_type = find_context.find_type
        else:
            self.__find_type = None
    
    @property
    def find_type(self):
        """
        @return Find type
        """
        return self.__find_type
    
    def update(self, updater: FindUpdater):
        """
        @param updater Find updater
        @return Updated context
        """
        res = copy.deepcopy(self)
        if updater.has_context_value("find_type"):
            res = res.with_find_type(updater.get_context_value("find_type"))
        if updater.has_context_value("update_root_container"):
            res = res.update_root_container(updater.get_context_value("update_root_container"))
        return res
    
    def update_root_container(self, do_update):
        """
        If needed, update the root container with container
        @param do_update If root has to be updated
        @return Same context but with given root
        """
        raise NotImplementedError("When using update_root_container method, the instantiated sub-class of FindContext must override it")

    def with_find_type(self, find_type):
        """
        Note: if wanted find type is the same, self instance is returned, else a new one is returned
        @param find_type Find type
        @return Same context but with given find type
        """
        if find_type == self.find_type:
            return self
        else:
            res = copy.deepcopy(self)
            res.find_type = find_type
            return res
    
    def get_find_type_container_description_prefix(self):
        """
        @return Prefix to container description associated to find type
        """
        if self.find_type == FindType.InParents:
            return "parents of "
        elif self.find_type == FindType.InChildren:
            return "children of "
        else:
            return ""

    def get_input_description(self, element=None):
        """
        @return Input description
        """
        if element is None:
            return "{NONE}"
        elif isinstance(element, ElementHolder):
            return f"[{element.description}]"
        else:
            return str(element)

    def get_input_complete_description(self, element=None):
        """
        @return Input complete description
        """
        if element is None:
            return "{NONE}"
        elif isinstance(element, ElementHolder):
            return f"[{element.complete_description}]"
        else:
            return str(element)

    def get_input_complete_description_and_details(self, element=None):
        """
        @return Input complete description and details
        """
        if element is None:
            return "{NONE}"
        elif isinstance(element, ElementHolder):
            return f"[{element.complete_description_and_details}]"
        else:
            return str(element)

    @staticmethod
    def default(find_type=None):
        """
        @return Default find context
        """
        if FindContext.__instance_default is None:
            FindContext.__instance_default = FindContext()
        res = FindContext.__instance_default
        
        if find_type is not None:
            res = res.with_find_type(find_type)
        
        return res
    
    
    
class WithRootContainerFindContext(FindContext):
    """
    Find context using containers.
    """

    def __init__(self, find_type=None, find_context=None):
        super().__init__(find_type, find_context)
        self.__root_container = None
        
        if isinstance(find_context, WithRootContainerFindContext):
            self.__root_container = find_context.root_container
        
    @property
    def root_container(self):
        """
        @return Root container
        """
        return self.__root_container
    
    def update(self, updater):
        res = super().update(updater)
        if updater.has_context_value("root_container"):
            res = res.with_root_container(updater.get_context_value("root_container"))
        return res
    
    def get_criteria_context(self):
        """
        @return Associated criteria context
        """
        return copy.deepcopy(self)
    
    def with_root_container(self, root_container):
        """
        Note: if wanted root is the same, self instance is returned, else a new one is returned
        @param __root_container Wanted root
        @return Same context but with given root
        """
        if root_container == self.__root_container:
            return self
        else:
            res = copy.deepcopy(self)
            res.__root_container = root_container
            return res
        
    
    
    
class ContainerFindContext(WithRootContainerFindContext):
    """
    Find context in a single container.
    """

    __instance_default = None

    def __init__(self, find_type=None, find_context=None):
        super().__init__(find_type, find_context)
        self.__container = None
        
        if isinstance(find_context, ContainerFindContext):
            self.__container = find_context.container
        
    @property
    def container(self):
        """
        @return Container
        """
        return self.__container
    
    def update(self, updater):
        res = super().update(updater)
        if updater.has_context_value("container"):
            res = res.with_container(updater.get_context_value("container"))
        return res
    
    def update_root_container(self, do_update):
        """
        If needed, update the root container with container
        @param do_update If root has to be updated
        @return Same context but with given root
        """
        if do_update and self.find_type != FindType.InParents:
            return self.with_root_container(self.container)
        else:
            return self
    
    def with_container(self, container, do_update=None):
        """
        Note 1: if wanted container is the same, self instance is returned, else a new one is returned
        Note 2: if root container is not already set, also set container as root
        @param container Wanted container
        @param do_update If root has to be updated
        @return Same context but with given container
        """
        if do_update is None and self.root_container is None:
            do_update = True
            
        if (container == self.container 
                and (not (do_update and self.find_type != FindType.InParents) or container == self.root_container) ):
            return self
        else:
            res = copy.deepcopy(self)
            res.__container = container
            return res.update_root_container(do_update)
        
    def get_input_description(self):
        return super().get_input_description(self.container)

    def get_input_complete_description(self):
        return super().get_input_complete_description(self.container)

    def get_input_complete_description_and_details(self):
        return super().get_input_complete_description_and_details(self.container)
    

    @staticmethod
    def default(find_type=None, container=None):
        """
        @return Default find context
        """
        if ContainerFindContext.__instance_default is None:
            ContainerFindContext.__instance_default = ContainerFindContext()
        res = ContainerFindContext.__instance_default
        
        if find_type is not None:
            res = res.with_find_type(find_type)
        if container is not None:
            res = res.with_container(container)
        
        return res
    
    
class ListContainersFindContext(WithRootContainerFindContext):
    """
    Find context in list of containers.
    """

    __instance_default = None

    def __init__(self, find_type=None, find_context=None):
        super().__init__(find_type, find_context)
        self.__containers = None
        self.__update_root_container_on_iteration = False
        
        if isinstance(find_context, ListContainersFindContext):
            self.__containers = find_context.containers
            self.__update_root_container_on_iteration = find_context.update_root_container_on_iteration
            
    def __iter__(self):
        self.__n = 0
        self.__find_builder = FindBuilder()
        return self

    def __next__(self):
        if self.__n < len(self.containers):
            res = self.__find_builder.context(find_context=self, container=self.containers[self.__n])
            self.__n += 1
            return res
        else:
            raise StopIteration
        
    @property
    def nb_containers(self):
        """
        @return Number of containers
        """
        if self.__containers is None:
            return 0
        else:
            return len(self.__containers)
        
    @property
    def containers(self):
        """
        @return Containers in which find element
        """
        return self.__containers
        
    @property
    def update_root_container_on_iteration(self):
        """
        @return If root container must be updated on iteration on containers
        """
        return self.__update_root_container_on_iteration
    
    def update(self, updater):
        res = super().update(updater)
        if updater.has_context_value("containers"):
            res = res.with_containers(updater.get_context_value("containers"))
        return res
    
    def update_root_container(self, do_update):
        """
        If needed, update the root container with container
        @param do_update If root has to be updated
        @return Same context but with given root
        """
        if do_update and self.find_type != FindType.InParents:
            return self.with_root_container(self.container)
        else:
            return self
    
    def with_containers(self, containers, do_update=None):
        """
        Note: if containers is not empty, returns a new context
        @param containers Wanted containers
        @param do_update If root has to be updated
        @return Same context but with given container
        """
        if do_update is None and self.root_container is None:
            do_update = True
            
        if containers == self.containers and do_update == self.update_root_container_on_iteration:
            return self
        else:
            res = copy.deepcopy(self)
            res.__containers = containers
            return res.with_update_root_container_on_iteration(do_update)
    
    def with_update_root_container_on_iteration(self, do_update=None):
        """
        If needed, update the root container on iteration
        @param do_update If root has to be updated
        @return Same context but with given container
        """
        if do_update == self.update_root_container_on_iteration:
            return self
        else:
            res = copy.deepcopy(self)
            res.__update_root_container_on_iteration = do_update
            return res
        
    def get_input_description(self):
        if self.containers is None or len(self.containers) == 0:
            return super().get_input_description(None)
        else:
            containers_descriptions = [super().get_input_description(c) for c in self.containers]
            return f"[{', '.join(containers_descriptions)}]"

    def get_input_complete_description(self):
        if self.containers is None or len(self.containers) == 0:
            return super().get_input_complete_description(None)
        else:
            containers_descriptions = [super().get_input_complete_description(c) for c in self.containers]
            return f"[{', '.join(containers_descriptions)}]"

    def get_input_complete_description_and_details(self):
        if self.containers is None or len(self.containers) == 0:
            return super().get_input_complete_description_and_details(None)
        else:
            containers_descriptions = [super().get_input_complete_description_and_details(c) for c in self.containers]
            return f"[{', '.join(containers_descriptions)}]"
    

    @staticmethod
    def default(find_type=None, containers=None):
        """
        @return Default find context
        """
        if ListContainersFindContext.__instance_default is None:
            ListContainersFindContext.__instance_default = ListContainersFindContext()
        res = ListContainersFindContext.__instance_default
        
        if find_type is not None:
            res = res.with_find_type(find_type)
        if containers is not None:
            res = res.with_containers(containers)
        
        return res
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
from enum import Enum


logger = logging.getLogger(__name__)


class DBCompareOperator(str, Enum):
    Different = ("!=", "Equal")
    Equal = ("==", "Different")
    Inferior = ("<", "SuperiorOrEqual")
    InferiorOrEqual = ("<=", "Superior")
    Superior = (">", "InferiorOrEqual")
    SuperiorOrEqual = (">=", "Inferior")
    In = ("in", "NotIn")
    NotIn = ("not in", "In")
    
    def __new__(cls, value, not_name):
        obj = str.__new__(cls, [value])
        obj._value_ = value
        obj.__not_name = not_name
        return obj
    
    @property
    def not_(self):
        return DBCompareOperator[self.__not_name]
    


class QueryBuilder():
    """
    Generic Query builder.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, name, db_client=None):
        self.__name = name
        self.__db_client = db_client
    
    @property
    def name(self):
        return self.__name
    
    @property
    def db_client(self):
        return self.__db_client
    
    @db_client.setter
    def db_client(self, client):
        self.__db_client = client
    
    def select(self, table_name, where_data: dict=None, where_compare_data: list=None, sql_return="*"):
        """
        Simple query & values builder of a select by filtering on given where data.
        @param where_data: dictionary of (field_name, value) for simple where clauses.
        @param where_compare_data: list of tuples (field_name, operator, value) for where clauses comparing fields with values.
        """
        raise NotImplementedError

    def insert(self, table_name, data: dict):
        """
        Simple query & values builder of an insert of given data.
        @param data: data to insert as dictionary of (field_name, value).
        """
        raise NotImplementedError
    
    def update(self, table_name, data: dict, where_data: dict=None, where_compare_data: list=None):
        """
        Simple query & values builder of an update of given data.
        @param data: data to update as dictionary of (field_name, value).
        @param where_data: dictionary of (field_name, value) for simple where clauses.
        @param where_compare_data: list of tuples (field_name, operator, value) for where clauses comparing fields with values.
        """
        raise NotImplementedError
    
    def delete(self, table_name, where_data: dict=None, where_compare_data: list=None):
        """
        Simple query & values builder of a delete by filtering on given where data.
        @param where_data: dictionary of (field_name, value) for simple where clauses.
        @param where_compare_data: list of tuples (field_name, operator, value) for where clauses comparing fields with values.
        """
        raise NotImplementedError
    
    def where(self, query, values, where_data: dict):
        """
        Add where clauses to current couple (query, values), and return a new couple (query, values).
        """
        raise NotImplementedError
    
    def where_compare(self, query, values, *, field_name=None, operator:DBCompareOperator=None, value=None, list_of_field_operator_value=None):
        """
        Add where clause with field value comparison to current couple (query, values), and return a new couple (query, values).
        To add multiple clauses in a single call, define list_of_field_operator_value as a list of tuples (field_name, operator, value).
        """
        if list_of_field_operator_value is not None:
            res_query, res_values = query, values
            for f_name, op, val in list_of_field_operator_value:
                res_query, res_values = self.where_compare(res_query, res_values, field_name=f_name, operator=op, value=val)
            return res_query, res_values
        else:
            raise NotImplementedError
        
    def where_in(self, query, values, field_name, field_values, not_in=False):
        """
        Add where in clause to current couple (query, values), and return a new couple (query, values).
        """
        raise NotImplementedError
        
    def where_is_null(self, query, values, field_name, is_not_null=False):
        """
        Add where is null clause to current couple (query, values), and return a new couple (query, values).
        """
        raise NotImplementedError
    
    def where_json_value(self, query, values, field_name, key, value, as_text_value=False):
        """
        Add where clause on a json field key value to current couple (query, values), and return a new couple (query, values).
        """
        raise NotImplementedError
        
    
    def to_sql(self, query):
        raise NotImplementedError
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools


logger = logging.getLogger(__name__)



class QueryManager():
    """
    Manage Query builders, agnostic to managed libraries.
    """
    
    def __init__(self, name):
        self.__name = name if name else "QueryManager"
        
        self.__builder_new_func_by_type = {}
        self.__default_builder_type = None
        
    def initialize(self):
        from holado_db.tools.db.query.pypika.pypika_query_builder import PypikaQueryBuilder
        if PypikaQueryBuilder.is_available():
            self.register_query_builder("pypika", PypikaQueryBuilder)
            self.__default_builder_type = "pypika"
        else:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug("PyPika package is not installed, this type of Query builder is not available")
        
    @property
    def name(self):
        return self.__name
        
    @property
    def default_builder_type(self):
        return self.__default_builder_type
        
    @default_builder_type.setter
    def default_builder_type(self, builder_type):
        self.__default_builder_type = builder_type
    
    def register_query_builder(self, builder_type, new_builder_func):
        self.__builder_new_func_by_type[builder_type] = new_builder_func
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Registered Query builder type '{builder_type}'")
    
    def new_default_query_builder(self, name, db_client):
        if self.__default_builder_type is None:
            raise TechnicalException(f"Default query builder type is not defined. Registered query builders: {list(self.__builder_new_func_by_type.keys())}")
        return self.new_query_builder(name, self.__default_builder_type, db_client)
    
    def new_query_builder(self, name, builder_type, db_client):
        if builder_type in self.__builder_new_func_by_type:
            try:
                res = self.__builder_new_func_by_type[builder_type](name, db_client)
            except Exception as exc:
                raise TechnicalException(f"Failed to create Query builder for builder type '{builder_type}'") from exc
        else:
            raise TechnicalException(f"Unmanaged Query builder type '{builder_type}'")
        return res
        
        
        
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import super
import logging
import os.path
from holado.common.context.context import Context
import threading
from holado.common.handlers.enums import ObjectStates
from holado.holado_config import Config
import importlib


logger = logging

def initialize_logger():
    global logger
    logger = logging.getLogger(__name__)


class SessionContext(Context):
    TSessionContext = None
    
    # Singleton management
    __instance = None
    __is_resetting_instance = False
    
    @staticmethod
    def instance() -> TSessionContext:
        # Note: If session context istance is under reset, consider it is already None
        if SessionContext.__is_resetting_instance:
            return None
        
        if SessionContext.__instance is None:
            # If needed, import session context module, and replace TSessionContext string by its class type
            if isinstance(SessionContext.TSessionContext, str):
                module_name, class_type = SessionContext.TSessionContext.rsplit('.', maxsplit=1)
                module = importlib.import_module(module_name)
                SessionContext.TSessionContext = getattr(module, class_type)
            
            # Create instance
            SessionContext.__instance = SessionContext.TSessionContext()
            logger.debug(f"Created session context of type {SessionContext.TSessionContext}")
            # logging.log(45, f"Created session context of type {SessionContext.TSessionContext}")
            # import traceback
            # logging.log(45, "".join(traceback.format_list(traceback.extract_stack())))
        return SessionContext.__instance
    
    @staticmethod
    def has_instance() -> bool:
        if SessionContext.__is_resetting_instance:
            return False
        else:
            return SessionContext.__instance is not None
    
    @staticmethod
    def _reset_instance():
        if SessionContext.__instance is not None:
            logger.debug(f"Resetting session context")
            SessionContext.__is_resetting_instance = True
            SessionContext.__instance = None
            SessionContext.__is_resetting_instance = False
            logger.debug(f"Reset of session context")
    
    def __init__(self, name="Session"):
        super().__init__(name, with_post_process=True)
        
        self.__with_session_path = None
        
        from holado.common.context.service_manager import ServiceManager
        self.__service_manager = ServiceManager(type(self))
        
        # Manage multitasking
        self.__multitask_lock = threading.RLock()
        
    def __del__(self):
        # Note: Override Object.__del__ since it supposes that SessionContext is initialized, whereas SessionContext is deleting
        try:
            if self.object_state in [ObjectStates.Deleting, ObjectStates.Deleted]:
                return
            
            self.delete_object()
        except Exception as exc:
            if "Python is likely shutting down" in str(exc):
                # Simply return
                return
            else:
                raise exc
    
    @property
    def _multitask_lock(self):
        return self.__multitask_lock
    
    def _delete_object(self):
        # if Tools.do_log(logger, logging.DEBUG):
        #     logger.debug("Interrupting and unregistering all threads for scenario [{}]".format(scenario.name))
        if self.has_threads_manager:
            logger.info(f"Delete session context - Interrupting and unregistering all threads...")
            self.threads_manager.interrupt_all_threads(scope="Session")
            self.threads_manager.unregister_all_threads(scope="Session", keep_alive=False)
        
        # Delete session context
        logger.info(f"Delete session context - Deleting context objects...")
        super()._delete_object()
        
    def configure(self, session_kwargs=None):
        """
        Override this method to configure the session context before new session creation.
        It is usually used to register new services.
        """
        if session_kwargs is None:
            session_kwargs = {}
            
        self.__with_session_path = session_kwargs.get("with_session_path", True)
        
        # Create this thread context
        self.multitask_manager.get_thread_context()
    
    def initialize(self, session_kwargs=None):
        """
        Override this method to initialize the session context after its configuration and new session creation.
        """
        pass
        
    @property
    def services(self):
        return self.__service_manager
    
    @property
    def with_session_path(self):
        return self.__with_session_path
    
    def get_object_getter_eval_string(self, obj, raise_not_found=True):
        from holado_python.standard_library.typing import Typing
        
        name = self.get_object_name(obj)
        if name is not None:
            if self.__service_manager.has_service(name):
                return f"{Typing.get_object_class_fullname(self)}.instance().{name}"
            else:
                return f"{Typing.get_object_class_fullname(self)}.instance().get_object('{name}')"
        
        if raise_not_found:
            from holado_core.common.exceptions.element_exception import ElementNotFoundException
            raise ElementNotFoundException(f"[{self.name}] Failed to find object of id {id(obj)}")
        else:
            return None

    def new_session(self, session_kwargs=None):
        from holado_python.common.tools.datetime import DateTime
        
        # Report session
        report_path = None
        if self.with_session_path:
            # Create new report path for this session
            name = "session_{}".format(DateTime.now(tz=Config.report_timezone).strftime("%Y-%m-%d_%H-%M-%S"))  # @UndefinedVariable
            report_path = self.path_manager.get_reports_path(name)
            logger.info(f"Reports location: {report_path}")
            print(f"Reports location: {report_path}")
        self.report_manager.new_session(report_path)
            
        # Logging configuration
        if self.with_session_path and SessionContext.instance().log_manager.in_file:
            log_filename = os.path.join(report_path, "logs", "report.log")
            self.path_manager.makedirs(log_filename)
            SessionContext.instance().log_manager.set_root_log_file(log_filename)
            SessionContext.instance().log_manager.set_config()
    


SessionContext.TSessionContext = SessionContext





# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.scenario.step_tools import StepTools
from holado.common.context.session_context import SessionContext
import logging
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_ais.ais.ais_messages import AISMessages
from holado_test.behave.behave import *  # @UnusedWildImport
from holado_ais.ais.enums import AISMessageType
from holado_core.common.tools.string_tools import StrTools
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from holado_core.common.block.function import Function
from holado_core.common.exceptions.functional_exception import FunctionalException

logger = logging.getLogger(__name__)


def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_text_interpreter():
    return __get_scenario_context().get_text_interpreter()

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()

def __get_ais_messages():
    return SessionContext.instance().ais_messages


if AISMessages.is_available():
    @Step(r"(?P<var_name>{Variable}) = decode AIS message of type (?P<msg_type>{Str}) from bitarray (?P<msg_bitarray>{Str})")
    def step_impl(context, var_name, msg_type, msg_bitarray):
        var_name = StepTools.evaluate_variable_name(var_name)
        msg_type = StepTools.evaluate_scenario_parameter(msg_type)
        msg_bitarray = StepTools.evaluate_scenario_parameter(msg_bitarray)
        
        res = __get_ais_messages().decode_message(msg_type, msg_bitarray)
        
        __get_variable_manager().register_variable(var_name, res)
    
    @Step(r"(?P<var_name>{Variable}) = encode AIS message")
    def step_impl(context, var_name):
        """
        Encode an AIS message to NMEA format
        """
        var_name = StepTools.evaluate_variable_name(var_name)
        table = BehaveStepTools.get_step_table(context)
        
        res = __get_ais_messages().encode(table)
        
        __get_variable_manager().register_variable(var_name, res)
    
    @Step(r"(?P<var_name>{Variable}) = encode AIS message (?P<message>{Any})")
    def step_impl(context, var_name, message):
        """
        Encode an AIS message to NMEA format
        """
        var_name = StepTools.evaluate_variable_name(var_name)
        msg = StepTools.evaluate_scenario_parameter(message)
        table = BehaveStepTools.get_step_table(context)
        
        if table is not None:
            kwargs = ValueTableConverter.convert_table_with_header_to_dict(table)
        else:
            kwargs = {}
        res = __get_ais_messages().encode(msg, **kwargs)
        
        __get_variable_manager().register_variable(var_name, res)
    
    @Step(r"(?P<var_name>{Variable}) = encode AIS raw payload (?P<raw_payload>{Any}) to NMEA")
    def step_impl(context, var_name, raw_payload):
        """
        Encode an AIS message to NMEA format
        """
        var_name = StepTools.evaluate_variable_name(var_name)
        raw_payload = StepTools.evaluate_scenario_parameter(raw_payload)
        table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
        
        if table is not None:
            kwargs = ValueTableConverter.convert_table_with_header_to_dict(table)
        else:
            kwargs = {}
        res = __get_ais_messages().encode_raw_payload(raw_payload, **kwargs)
        
        __get_variable_manager().register_variable(var_name, res)
    
    @Step(r"(?P<var_name>{Variable}) = new AIS message of type (?P<ais_message_type>{Str}) as bitarray bytes")
    def step_impl(context, var_name, ais_message_type):
        var_name = StepTools.evaluate_variable_name(var_name)
        msg_type = StepTools.evaluate_scenario_parameter(ais_message_type)
        
        if hasattr(context, 'table') and context.table is not None:
            table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
        else:
            table = {}
        
        res = __get_ais_messages().new_message_as_bitarray_bytes(msg_type, table)
        
        __get_variable_manager().register_variable(var_name, res)
        
    @Step(r"(?P<var_name>{Variable}) = new AIS message of type (?P<msg_type>{Int})")
    def step_impl(context, var_name, msg_type:int):
        var_name = StepTools.evaluate_variable_name(var_name)
        msg_type = StepTools.evaluate_scenario_parameter(msg_type)
        table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
        
        res = __get_ais_messages().new_message(msg_type, table)
        
        __get_variable_manager().register_variable(var_name, res)

    @Step(r"(?P<var_name>{Variable}) = convert AIS message (?P<msg>{Str}) to binary string")
    def step_impl(context, var_name, msg):
        var_name = StepTools.evaluate_variable_name(var_name)
        msg = StepTools.evaluate_scenario_parameter(msg)
        
        res = __get_ais_messages().convert_message_to_binary_str(msg)
        
        __get_variable_manager().register_variable(var_name, res)

    @Step(r"(?P<var_name>{Variable}) = convert AIS message (?P<msg>{Str}) to bytes")
    def step_impl(context, var_name, msg):
        var_name = StepTools.evaluate_variable_name(var_name)
        msg = StepTools.evaluate_scenario_parameter(msg)
        
        res = __get_ais_messages().convert_message_to_bytes(msg)
        
        __get_variable_manager().register_variable(var_name, res)

    @Step(r"(?P<var_name>{Variable}) = convert AIS message (?P<msg>{Str}) to hexadecimal string")
    def step_impl(context, var_name, msg):
        var_name = StepTools.evaluate_variable_name(var_name)
        msg = StepTools.evaluate_scenario_parameter(msg)
        
        res = __get_ais_messages().convert_message_to_bytes(msg)
        res = StrTools.to_hex(res)
        
        __get_variable_manager().register_variable(var_name, res)
    
    @Step(r"(?:(?P<tagblock_var_name>{Variable}), )?(?P<var_name>{Variable}) = decode NMEA AIS raw (?P<ais_raw>{Bytes}|{Str}) as dictionary(?: \((?P<merge_tag_block_and_message_dicts_str>merge tag block and message dictionaries)?\))?")
    def step_impl(context, tagblock_var_name, var_name, ais_raw, merge_tag_block_and_message_dicts_str):
        var_name = StepTools.evaluate_variable_name(var_name)
        ais_raw = StepTools.evaluate_scenario_parameter(ais_raw)
        merge_tag_block_and_message_dicts = merge_tag_block_and_message_dicts_str is not None
        
        res = __get_ais_messages().decode_nmea_raw_as_dict(ais_raw, merge_tag_block_and_message_dicts=merge_tag_block_and_message_dicts)
        
        if tagblock_var_name is None:
            __get_variable_manager().register_variable(var_name, res)
        else:
            if len(res) == 2:
                tb_dict, msg_dict = res
            else:
                tb_dict, msg_dict = None, res
            __get_variable_manager().register_variable(tagblock_var_name, tb_dict)
            __get_variable_manager().register_variable(var_name, msg_dict)
    
    @Step(r"(?P<var_name>{Variable}) = decode NMEA AIS message (?P<encoded_ais_message>{Str})(?P<dictionary_str> as dictionary)?")
    def step_impl(context, var_name, encoded_ais_message, dictionary_str):
        var_name = StepTools.evaluate_variable_name(var_name)
        encoded_ais_message = StepTools.evaluate_scenario_parameter(encoded_ais_message)
        
        if isinstance(encoded_ais_message, list):
            res = __get_ais_messages().decode(*encoded_ais_message)
        else:
            res = __get_ais_messages().decode(encoded_msg=encoded_ais_message)
            
        if dictionary_str is not None:
            res = res.asdict()
            
        __get_variable_manager().register_variable(var_name, res)
    
    #TODO EKL: rename step
    @Given(r"(?P<var_name>{Variable}) = split NMEA AIS message (?P<ais_message>{Bytes}|{Str}) to fields")
    def step_impl(context, var_name, ais_message):
        var_name = StepTools.evaluate_variable_name(var_name)
        ais_message = StepTools.evaluate_scenario_parameter(ais_message)
        
        res = __get_ais_messages().split_message_to_fields(ais_message)
        
        __get_variable_manager().register_variable(var_name, res)
    
    @Given(r"(?P<var_name>{Variable}) = split (?P<ais_raw>{Bytes}|{Str}) to NMEA AIS sentences")
    def step_impl(context, var_name, ais_raw):
        var_name = StepTools.evaluate_variable_name(var_name)
        ais_raw = StepTools.evaluate_scenario_parameter(ais_raw)
        
        res = __get_ais_messages().split_raw_to_sentences(ais_raw)
        
        __get_variable_manager().register_variable(var_name, res)
        
    @Given(r"(?P<tb_var_name>{Variable}), (?P<msg_var_name>{Variable}) = split NMEA AIS sentence (?P<ais_sentence>{Bytes}|{Str}) to tag block and message")
    def step_impl(context, tb_var_name, msg_var_name, ais_sentence):
        tb_var_name = StepTools.evaluate_variable_name(tb_var_name)
        msg_var_name = StepTools.evaluate_variable_name(msg_var_name)
        ais_sentence = StepTools.evaluate_scenario_parameter(ais_sentence)
        
        tb, msg = __get_ais_messages().split_sentence_to_tag_block_and_message(ais_sentence)
        
        __get_variable_manager().register_variable(tb_var_name, tb)
        __get_variable_manager().register_variable(msg_var_name, msg)
        
    @Given(r"(?P<var_name>{Variable}) = split NMEA AIS Tag Block (?P<tag_block>{Bytes}|{Str}) to dict")
    def step_impl(context, var_name, tag_block):
        var_name = StepTools.evaluate_variable_name(var_name)
        tag_block = StepTools.evaluate_scenario_parameter(tag_block)
        
        res = __get_ais_messages().convert_tag_block_to_dict(tag_block)
        
        __get_variable_manager().register_variable(var_name, res)
        
    @Given(r"(?P<var_name>{Variable}) = NMEA AIS message (?P<ais_message>{Bytes}|{Str}) with message ID replaced by (?P<replacement_id>{Bytes}|{Str})")
    def step_impl(context, var_name, ais_message, replacement_id):
        var_name = StepTools.evaluate_variable_name(var_name)
        ais_message = StepTools.evaluate_scenario_parameter(ais_message)
        replacement_id = StepTools.evaluate_scenario_parameter(replacement_id)
        
        res = __get_ais_messages().replace_message_id_in_message(ais_message, replacement_id)
        
        __get_variable_manager().register_variable(var_name, res)
    
    





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools
from holado_core.common.criterias.tools.criteria_context import CriteriaContext,\
    ContainerCriteriaContext
from holado_core.common.criterias.tools.criteria_parameters import CriteriaParameters
from holado.holado_config import Config
import logging
import abc

logger = logging.getLogger(__name__)



class Criteria(object):
    """ Base class of a family used to validate elements.
    """
    __metaclass__ = abc.ABCMeta
    
    def validate(self, element, criteria_context: CriteriaContext, criteria_parameters: CriteriaParameters):
        """ Verify if element verifies expected criteria.
        If element is None and criteria_context is a ContainerCriteriaContext, criteria_context.container is used as element
        @param element Element to validate
        @param criteria_context Criteria context
        @param criteria_parameters Criteria parameters
        @return True if given element verifies expected criteria
        """
        if element is None:
            if isinstance(criteria_context, ContainerCriteriaContext) and criteria_context.container is not None:
                return self.validate(criteria_context.container, criteria_context, criteria_parameters)
            else:
                raise TechnicalException("Find context has no container")
            
        # Prepare analyze time spent
        if criteria_parameters.analyze_time_spent:
            start = Tools.timer_s()
        
        res = self.validate_element(element, criteria_context, criteria_parameters)
        
        # Analyze time spent
        if criteria_parameters.analyze_time_spent:
            duration = Tools.timer_s() - start
            if duration > Config.threshold_warn_time_spent_s:
                logger.warning(f"{self._get_indent_string_level(criteria_parameters)}[{self}].validate({element}) -> {res} (took {duration} s)")
        
        return res

    def validate_element(self, element, criteria_context: CriteriaContext, criteria_parameters: CriteriaParameters):
        return True

    def __str__(self):
        return self.__class__.__name__

    def reset(self):
        """ Reset the criteria."""
        pass
    
    def _get_indent_string_level(self, criteria_parameters: CriteriaParameters=None, level=None):
        if level is not None:
            return Tools.get_indent_string(level * 4)
        elif criteria_parameters is not None:
            return self._get_indent_string_level(level=criteria_parameters.find_level)
        else:
            raise TechnicalException("level or criteria_parameters must be defined")
    
    
    


#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.exceptions.technical_exception import TechnicalException
import io



class Converter(object):
    
    @classmethod
    def is_boolean(cls, value):
        return isinstance(value, bool)\
            or isinstance(value, str) and (value == "False" or value == "True")

    @classmethod
    def to_boolean(cls, value):
        if isinstance(value, bool):
            return value
        elif isinstance(value, str):
            if value == "True":
                return True
            elif value == "False":
                return False
            else:
                raise TechnicalException(f"String value '{value}' is not a boolean")
        else:
            from holado_python.standard_library.typing import Typing
            raise TechnicalException(f"Value [{value}] (type: {Typing.get_object_class_fullname(value)}) is not a boolean")

    @classmethod
    def is_dict(cls, obj):
        return isinstance(obj, dict)

    @classmethod
    def to_dict(cls, value):
        if isinstance(value, dict):
            return value
        else:
            return dict(value)

    @classmethod
    def is_list(cls, obj):
        return isinstance(obj, list)

    @classmethod
    def to_list(cls, value):
        if value is None:
            return None
        elif isinstance(value, list):
            return value
        elif cls.is_iterable(value):
            return list(value)
        else:
            from holado_python.standard_library.typing import Typing
            raise TechnicalException(f"Unmanaged convertion to list of object of type {Typing.get_object_class_fullname(value)}")

    @classmethod
    def is_integer(cls, value):
        try:
            int(value)
        except ValueError:
            return False
        else:
            return True

    @classmethod
    def to_integer(cls, value):
        if isinstance(value, int):
            return value
        else:
            return int(value)

    @classmethod
    def is_float(cls, value):
        try:
            float(value)
        except ValueError:
            return False
        else:
            return True

    @classmethod
    def to_float(cls, value):
        if isinstance(value, float):
            return value
        else:
            return float(value)
        
    @classmethod
    def is_primitive(cls, obj):
        from holado_python.standard_library.typing import Typing
        return (not hasattr(obj, '__dict__') or type(obj) == type) and not Typing.is_NamedTuple(obj)
        
    @classmethod
    def is_iterable(cls, obj):
        try:
            _ = (e for e in obj)
        except TypeError:
            return False
        else:
            return True
    
    @classmethod
    def is_file_like(cls, obj):
        return isinstance(obj, io.TextIOBase) or isinstance(obj, io.BufferedIOBase) or isinstance(obj, io.RawIOBase) or isinstance(obj, io.IOBase)



